from busbus.util import Config

import os
import pytest


def test_config_defaults():
    config = Config()
    assert config['url_cache_dir'].endswith('/.busbus/cache')


def test_config_without_home():
    old_home = os.environ['HOME']
    del os.environ['HOME']
    config = Config()
    assert (os.path.abspath(config['url_cache_dir']) ==
            os.path.abspath('.busbus/cache'))
    os.environ['HOME'] = old_home


def test_config_busbus_dir():
    config = Config({'busbus_dir': '/path/to/.busbus'})
    assert config['url_cache_dir'] == '/path/to/.busbus/cache'


def test_config_keyerror():
    config = Config()
    with pytest.raises(KeyError) as exc:
        config['The weather in London']
    assert exc.value.args[0] == 'The weather in London'



import busbus.provider
from busbus import util

import arrow
import collections
import json
from six.moves import reduce


class LazyEntityProperty(object):

    def __init__(self, f, *args, **kwargs):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.f(*self.args, **self.kwargs)


class BaseEntity(collections.Mapping):
    __repr_attrs__ = ('id',)
    __derived__ = False

    def __init__(self, provider, **kwargs):
        self.provider = provider
        self._lazy_properties = {}

        for attr in getattr(self, '__attrs__', []):
            if isinstance(kwargs.get(attr, None), LazyEntityProperty):
                self._lazy_properties[attr] = kwargs[attr]
            else:
                setattr(self, attr, kwargs.get(attr, None))

        if not self.__derived__:
            provider._new_entity(self)

    def __repr__(self):
        return u'<{0}({1})>'.format(
            util.clsname(self), ','.join(
                '{0}={1!r}'.format(i, getattr(self, i))
                for i in self.__repr_attrs__))

    def __getattr__(self, name):
        if name in self._lazy_properties:
            value = self._lazy_properties[name]()
            del self._lazy_properties[name]
            setattr(self, name, value)
            return value
        if '.' in name:  # nested attribute
            return reduce(getattr, name.split('.'), self)
        raise AttributeError(name)

    def __getitem__(self, name):
        try:
            return getattr(self, name)
        except (AttributeError, TypeError):
            raise KeyError(name)

    def __iter__(self):
        yield 'provider'
        for attr in self.__attrs__:
            if getattr(self, attr, None) is not None:
                yield attr

    def __len__(self):
        return 1 + sum([1 for attr in self.__attrs__
                        if getattr(self, attr, None) is not None])


class BaseEntityJSONEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, (BaseEntity, busbus.provider.ProviderBase)):
            return dict(o)
        elif isinstance(o, arrow.Arrow):
            return o.timestamp
        elif isinstance(o, collections.Iterable):
            return list(o)
        return super(BaseEntityJSONEncoder, self).default(o)



from __future__ import absolute_import

from busbus import util

import csv
import io
import six


class CSVReader(util.Iterable):
    """
    Ordinarily one would use csv.reader for their day-to-day CSV file reading.
    However, the csv module expects a str object in both Python 2 and 3; of
    course, this means that it wants bytes in Python 2 and unicode in Python 3.
    Likewise, the data it spits out is a str in both languages, but we want the
    equivalent of unicode in both languages (unicode and str, respectively).
    """

    def __init__(self, csvfile):
        if six.PY3:
            csvfile = io.TextIOWrapper(csvfile, encoding='utf-8')
        self.reader = csv.reader(csvfile)
        self.header = next(self.reader)
        if six.PY2:
            self.header = [x.decode('utf-8') for x in self.header]

    def __next__(self):
        row = next(self.reader)
        if six.PY2:
            func = lambda x: x.decode('utf-8') if x else None
        else:
            func = lambda x: x if x else None
        return [func(x.strip()) for x in row]



import busbus
from busbus.provider import ProviderBase
from busbus.provider.gtfs import GTFSMixin

import arrow
import contextlib
try:
    # prefer the stdlib version of mock (>= 3.3)
    import unittest.mock as mock
except ImportError:
    import mock
import os
from pkg_resources import resource_listdir, resource_string
import pytest
import responses
import six
import zipfile


def mock_gtfs_zip(name):
    path = os.path.join('data', name, 'gtfs')
    with contextlib.closing(six.BytesIO()) as zipdata:
        with zipfile.ZipFile(zipdata, 'w') as z:
            for filename in resource_listdir(__name__, path):
                s = resource_string(__name__, os.path.join(path, filename))
                z.writestr(filename, s)
        return zipdata.getvalue()


class SampleGTFSProvider(GTFSMixin, ProviderBase):
    gtfs_url = ('https://developers.google.com/transit/gtfs/examples/'
                'sample-feed.zip')

    def __init__(self, engine=None):
        # https://developers.google.com/transit/gtfs/examples/gtfs-feed
        # FIXME We should eventually roll our own GTFS feed as well in order to
        # test everything
        super(SampleGTFSProvider, self).__init__(engine, self.gtfs_url)

    @property
    def arrivals(self):
        # Set a current time that fits within the sample feed's dates
        with mock.patch('arrow.now') as mock_now:
            mock_now.return_value = arrow.get('2007-06-03T06:45:00-07:00')
            return super(SampleGTFSProvider, self).arrivals


@pytest.fixture(scope='session')
def engine_config():
    return {
        'gtfs_db_path': ':memory:',
    }


@pytest.fixture(scope='session')
def engine(engine_config):
    return busbus.Engine(engine_config)


@pytest.fixture(scope='session')
def gtfs_zip_data():
    return mock_gtfs_zip('_sample')


@pytest.fixture(scope='session')
@responses.activate
def provider(engine, gtfs_zip_data):
    responses.add(responses.GET, SampleGTFSProvider.gtfs_url,
                  body=gtfs_zip_data, status=200,
                  content_type='application/zip')
    return SampleGTFSProvider(engine)



# coding=utf-8

from busbus.provider import ProviderBase
from busbus.queryable import Queryable
from .conftest import SampleGTFSProvider, mock_gtfs_zip

import pytest
import responses
import requests
from wsgi_intercept import requests_intercept, add_wsgi_intercept

web = pytest.importorskip('busbus.web')
cherrypy = pytest.importorskip('cherrypy')


class DumbUselessProvider(ProviderBase):

    def get(self, entity, id, default=None):
        return default

    @property
    def agencies(self):
        return Queryable(())

    @property
    def stops(self):
        return Queryable(())

    @property
    def routes(self):
        return Queryable(())

    @property
    def arrivals(self):
        return Queryable(())


@pytest.fixture(scope='module')
def web_engine(engine_config):
    engine = web.Engine(engine_config)
    responses.add(responses.GET, SampleGTFSProvider.gtfs_url,
                  body=mock_gtfs_zip('_sample'), status=200,
                  content_type='application/zip')
    SampleGTFSProvider(engine)
    DumbUselessProvider(engine)

    return engine


@pytest.fixture(scope='module')
@responses.activate
def url_prefix(request, web_engine):
    # https://cherrypy.readthedocs.org/en/latest/deploy.html
    # § Embedding into another WSGI framework
    cherrypy.config.update({'environment': 'embedded',
                            'global': {'request.throw_errors': True}})
    cherrypy.server.unsubscribe()  # disables built-in HTTP server
    cherrypy.engine.start()

    # intercept requests
    host = 'busbus.invalid'
    port = 8080
    requests_intercept.install()
    add_wsgi_intercept(host, port, lambda: cherrypy.tree.mount(web_engine))

    @request.addfinalizer
    def fin():
        requests_intercept.uninstall()
        cherrypy.engine.exit()

    return 'http://{0}:{1}/'.format(host, port)


@pytest.fixture(scope='module')
def provider_id(web_engine):
    for id, provider in web_engine._providers.items():
        if isinstance(provider, SampleGTFSProvider):
            return id


@pytest.fixture(scope='module')
def dumb_provider_id(web_engine):
    for id, provider in web_engine._providers.items():
        if isinstance(provider, DumbUselessProvider):
            return id


def get(url, code=200):
    resp = requests.get(url)
    assert resp.status_code == code
    assert resp.headers['content-type'] == 'application/json'
    data = resp.json()
    assert 'request' in data
    assert 'status' in data['request']
    assert data['request']['status'] in (('error',) if code // 100 == 4
                                         else ('ok', 'help'))
    return (data, resp)


def test_index(url_prefix):
    data, resp = get(url_prefix)
    assert data['request']['status'] == 'help'
    for attr in ('_entities', '_actions'):
        assert isinstance(data.get(attr, None), list)


def test_invalid_entity(url_prefix):
    get(url_prefix + 'invalid_entity', 404)


def test_invalid_action(url_prefix):
    get(url_prefix + 'providers/invalid_action', 404)


def test_nested(web_engine, url_prefix):
    data, resp = get(url_prefix + ('arrivals?stop.name=Bullfrog (Demo)&'
                                   'start_time=2007-06-03T06:45:00-07:00'))
    assert all(a['stop']['id'] == 'BULLFROG' for a in data['arrivals'])


def test_query(url_prefix):
    data, resp = get(url_prefix + 'stops')
    assert 'stops' in data
    assert len(data['stops']) == 9


def test_unexpand_none(url_prefix):
    data, resp = get(url_prefix + 'routes')
    assert set(data['routes'][0]['agency'].keys()) == set(['id'])


def test_unexpand_dict():
    assert web.unexpand({1: 2}, ()) == {1: 2}


def test_unexpand_agencies(url_prefix):
    data, resp = get(url_prefix + 'routes?_expand=agencies')
    assert 'timezone' in data['routes'][0]['agency']


def test_limit(url_prefix):
    data, resp = get(url_prefix + 'routes?_limit=1')
    assert len(data['routes']) == 1


def test_invalid_limit(url_prefix):
    get(url_prefix + 'routes?_limit=-422', 422)


def test_limit_on_action(url_prefix):
    data, resp = get(url_prefix + ('stops/find?latitude=36.914778&'
                                   'longitude=-116.767900&distance=10000&'
                                   '_limit=1'))
    assert len(data['stops']) == 1


def test_stops_find(url_prefix):
    data, resp = get(url_prefix + ('stops/find?latitude=36.914778&'
                                   'longitude=-116.767900&distance=100'))
    assert len(data['stops']) == 1
    assert data['stops'][0]['id'] == 'NADAV'


def test_stops_find_missing_attrs(url_prefix):
    data, resp = get(url_prefix + 'stops/find', 422)
    assert data['error'].startswith('missing attributes')


def test_routes_directions(url_prefix, provider_id):
    data, resp = get(url_prefix + ('routes/directions?route.id=AB&'
                                   'provider.id={0}'.format(provider_id)))
    assert len(data['directions']) == 2


def test_routes_directions_missing_attrs(url_prefix):
    data, resp = get(url_prefix + 'routes/directions', 422)
    assert data['error'].startswith('missing attributes')


def test_arrivals_realtime_true(url_prefix, provider_id):
    data, resp = get(url_prefix + ('arrivals?stop.id=AMV&realtime=TRUE&'
                                   'start_time=2007-06-03T06:45:00-07:00&'
                                   'provider.id={0}'.format(provider_id)))
    assert data['request']['params']['realtime'] == True
    assert len(data['arrivals']) == 0


def test_arrivals_realtime_false(url_prefix, provider_id):
    data, resp = get(url_prefix + ('arrivals?stop.id=AMV&realtime=no&'
                                   'start_time=2007-06-03T06:45:00-07:00&'
                                   'provider.id={0}'.format(provider_id)))
    assert data['request']['params']['realtime'] == False
    assert len(data['arrivals']) == 1


def test_arrivals_realtime_invalid(url_prefix, provider_id):
    data, resp = get(url_prefix + ('arrivals?stop.id=AMV&realtime=butts&'
                                   'start_time=2007-06-03T06:45:00-07:00&'
                                   'provider.id={0}'.format(provider_id)), 422)


def test_arrivals_provider_id(url_prefix, dumb_provider_id):
    data, resp = get(url_prefix + ('arrivals?provider.id={0}'
                                   .format(dumb_provider_id)))
    assert len(data['arrivals']) == 0


def test_arrivals_invalid_provider_id(url_prefix):
    data, resp = get(url_prefix + 'arrivals?provider.id=butts')
    assert len(data['arrivals']) == 0



from busbus.queryable import Queryable

from six.moves import range
import pytest


def test_queryable_where_func():
    q = Queryable(range(10)).where(lambda x: x % 5 == 0)
    assert next(q) == 0
    assert next(q) == 5
    with pytest.raises(StopIteration):
        next(q)


def test_queryable_where_manyfunc():
    q = Queryable(range(10)).where(lambda x: x % 2 == 0)
    q = q.where(lambda x: x % 3 == 0)
    assert next(q) == 0
    assert next(q) == 6
    with pytest.raises(StopIteration):
        next(q)


def test_queryable_where_kwargs(provider):
    q = provider.agencies.where(id='DTA')
    assert len(list(q)) == 1
    q = provider.agencies.where(id='The weather in London')
    with pytest.raises(StopIteration):
        next(q)


def test_queryable_where_none():
    q = Queryable(range(10))
    q_prime = q.where()
    assert q is q_prime


def test_queryable_dict_obj_kwargs():
    q = Queryable({'id': x, 'odd': bool(x % 2)} for x in range(10))
    assert next(q.where(odd=True))['id'] == 1


def test_queryable_unknown_obj_kwargs():
    q = Queryable(range(10)).where(qwerty='uiop')
    with pytest.raises(StopIteration):
        next(q)


@pytest.fixture(scope='function')
def qchain():
    q1 = Queryable(range(10)).where(lambda x: x % 3 == 0)
    q2 = Queryable(range(10)).where(lambda x: x % 2 == 0)
    return Queryable.chain(q1, q2)


def test_chained_len(qchain):
    assert len(list(qchain)) == 9


def test_chained_set(qchain):
    assert set(qchain) == set([0, 2, 3, 4, 6, 8, 9])


def test_chained_where_len(qchain):
    assert len(list(qchain.where(lambda x: x % 6 == 0))) == 4


def test_chained_where_set(qchain):
    assert set(qchain.where(lambda x: x % 6 == 0)) == set([0, 6])



import busbus
from busbus.entity import BaseEntityJSONEncoder
from busbus.provider import ProviderBase
from busbus.queryable import Queryable

import cherrypy
import collections
import itertools
import types


def json_handler(*args, **kwargs):
    value = cherrypy.serving.request._json_inner_handler(*args, **kwargs)
    return BaseEntityJSONEncoder().encode(value).encode('utf-8')
cherrypy.config['tools.json_out.handler'] = json_handler


EXPAND_TYPES = {
    'providers': ProviderBase,
    'agencies': busbus.Agency,
    'stops': busbus.Stop,
    'routes': busbus.Route,
    'arrivals': busbus.Arrival,
}


def unexpand_init(result, to_expand):
    return ({attr: unexpand(value, to_expand)
             for attr, value in dict(obj).items()}
            for obj in result)


def unexpand(obj, to_expand):
    for name, cls in EXPAND_TYPES.items():
        if isinstance(obj, cls):
            if name not in to_expand:
                return {'id': obj.id}
            else:
                return {attr: unexpand(value, to_expand)
                        for attr, value in dict(obj).items()}
    if isinstance(obj, dict):
        return {attr: unexpand(value, to_expand)
                for attr, value in obj.items()}
    if isinstance(obj, (list, tuple, collections.Iterator)):
        return (unexpand(value, to_expand) for value in obj)
    return obj


class APIError(Exception):

    def __init__(self, msg, error_code=500):
        self.msg = msg
        self.error_code = error_code


class EndpointNotFoundError(APIError):

    def __init__(self, entity, action=None):
        super(EndpointNotFoundError, self).__init__(
            'Endpoint /{0} not found'.format(
                entity + '/' + action if action else entity), 404)


class Engine(busbus.Engine):

    def __init__(self, *args, **kwargs):
        # perhaps fix this to use a decorator somehow?
        self._entity_actions = {
            ('stops', 'find'): (self.stops_find, 'stops'),
            ('routes', 'directions'): (self.routes_directions, 'directions'),
        }
        super(Engine, self).__init__(*args, **kwargs)

    @cherrypy.popargs('entity', 'action')
    @cherrypy.expose
    @cherrypy.tools.json_out()
    def default(self, entity=None, action=None, **kwargs):
        if entity is None:
            return self.help()

        response = {
            'request': {
                'status': 'ok',
                'entity': entity,
                'params': kwargs,
            }
        }

        try:
            to_expand = (kwargs.pop('_expand').split(',')
                         if '_expand' in kwargs else [])
            if to_expand:
                response['request']['expand'] = to_expand

            limit = kwargs.pop('_limit', None)
            if limit:
                try:
                    limit = int(limit)
                    if limit <= 0:
                        raise ValueError()
                except ValueError:
                    raise APIError('_limit must be a positive integer', 422)
                response['request']['limit'] = limit

            if 'realtime' in kwargs:
                if kwargs['realtime'] in ('y', 'Y', 'yes', 'Yes', 'YES',
                                          'true', 'True', 'TRUE',
                                          'on', 'On', 'ON'):
                    kwargs['realtime'] = True
                elif kwargs['realtime'] in ('n', 'N', 'no', 'No', 'NO',
                                            'false', 'False', 'FALSE',
                                            'off', 'Off', 'OFF'):
                    kwargs['realtime'] = False
                else:
                    raise APIError('realtime is not a boolean', 422)

            if action:
                response['request']['action'] = action
                if (entity, action) in self._entity_actions:
                    func, entity = self._entity_actions[(entity, action)]
                    result = func(**kwargs)
                else:
                    raise EndpointNotFoundError(entity, action)
            else:
                if 'provider.id' in kwargs:
                    provider_id = kwargs.pop('provider.id')
                    if provider_id in self._providers:
                        provider = self._providers[provider_id]
                        entity_func = getattr(provider, entity, None)
                    else:
                        entity_func = Queryable(())
                else:
                    entity_func = getattr(self, entity, None)
                if entity_func is not None:
                    result = entity_func.where(**kwargs)
                else:
                    raise EndpointNotFoundError(entity)

            if limit:
                result = itertools.islice(result, limit)

            response[entity] = unexpand_init(result, to_expand)
        except APIError as exc:
            response['request']['status'] = 'error'
            response['error'] = exc.msg
            cherrypy.response.status = exc.error_code

        return response

    def help(self):
        return {
            'request': {
                'status': 'help',
            },
            '_entities': EXPAND_TYPES.keys(),
            '_actions': self._entity_actions.keys(),
        }

    def stops_find(self, **kwargs):
        expected = ('latitude', 'longitude', 'distance')
        if all(x in kwargs for x in expected):
            for x in expected:
                kwargs[x] = float(kwargs[x])
            latlon = (kwargs['latitude'], kwargs['longitude'])
            return super(Engine, self).stops.where(
                lambda s: s.distance_to(latlon) <= kwargs['distance'])
        else:
            raise APIError('missing attributes: ' + ','.join(
                x for x in expected if x not in kwargs), 422)

    def routes_directions(self, **kwargs):
        expected = ('route.id', 'provider.id')
        missing = [x for x in expected if x not in kwargs]
        if missing:
            raise APIError('missing attributes: ' + ','.join(missing), 422)
        provider = self._providers[kwargs['provider.id']]
        route = provider.get(busbus.Route, kwargs['route.id'])
        return route.directions



from abc import ABCMeta, abstractmethod
import collections
from datetime import datetime, timedelta
import math
import os
import requests
import six
import time

import busbus.entity


@six.add_metaclass(ABCMeta)
class Iterable(object):

    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self):
        """Provide the next item from the iterator in this method."""

    # Python 2 compatibility
    def next(self):
        return self.__next__()


class Config(collections.defaultdict):

    def __init__(self, config=None):
        if config:
            for k, v in config.items():
                self[k] = v
        super(Config, self).__init__()

    def __missing__(self, key):
        if key == 'busbus_dir':
            if os.getenv('HOME'):
                return os.path.join(os.getenv('HOME'), '.busbus')
            else:
                return os.path.join(os.getcwd(), '.busbus')
        elif key == 'url_cache_dir':
            return os.path.join(self['busbus_dir'], 'cache')
        elif key == 'gtfs_db_path':
            return os.path.join(self['busbus_dir'], 'gtfs.sqlite3')
        else:
            raise KeyError(key)


class RateLimitRequests(requests.Session):
    """
    requests.Session subclass that implements rate limiting, both for all
    requests on this session as well as separate rate limits for each
    individual request.
    """

    def __init__(self, interval=0, url_interval=0):
        self._cache = {}
        self._last_request = datetime.min
        self._interval = timedelta(seconds=interval)
        self._url_interval = timedelta(seconds=url_interval)

        super(RateLimitRequests, self).__init__()

    def request(self, *args, **kwargs):
        key = freezehash((args, kwargs))
        if key in self._cache:
            t, resp = self._cache[key]
            if datetime.now() - t <= self._url_interval:
                return resp
        time.sleep(max((self._last_request + self._interval -
                        datetime.now()).total_seconds(), 0))
        resp = super(RateLimitRequests, self).request(*args, **kwargs)
        t = datetime.now()
        self._cache[key] = (t, resp)
        self._last_request = t
        return resp


def entity_type(obj):
    """Return the type just above BaseEntity in method resolution order."""
    if not isinstance(obj, type):
        obj = type(obj)
    mro = obj.mro()
    if busbus.entity.BaseEntity in mro:
        index = mro.index(busbus.entity.BaseEntity)
        if index == 0:
            raise TypeError(obj)
        else:
            return mro[index - 1]
    else:
        raise TypeError(obj)


def clsname(obj):
    return '{0}.{1}'.format(type(obj).__module__, type(obj).__name__)


def freezehash(obj):
    if isinstance(obj, (dict, collections.Mapping)):
        return hash(frozenset((k, freezehash(v)) for k, v in obj.items()))
    elif isinstance(obj, (tuple, list, set)):
        return hash(frozenset(freezehash(x) for x in obj))
    else:
        return hash(obj)


def dist(lat1, lon1, lat2, lon2):
    """
    Returns the distance between two latitude/longitude pairs in
    meters.
    """
    lat1, lon1, lat2, lon2 = map(math.radians, (lat1, lon1, lat2, lon2))
    return math.acos(math.sin(lat1) * math.sin(lat2) +
                     math.cos(lat1) * math.cos(lat2) *
                     math.cos(abs(lon2 - lon1))) * 6371000



from .conftest import SampleGTFSProvider

import busbus
from busbus.provider import ProviderBase
from busbus.provider.gtfs import SQLEntityMixin
from busbus.util import Config

import arrow
from collections import OrderedDict
import datetime
import mock
import pytest
import responses
import six


def test_provider_without_engine():
    class TestProvider(ProviderBase):

        def get(self, *args, **kwargs):
            raise NotImplementedError

        @property
        def agencies(self):
            raise NotImplementedError

        @property
        def routes(self):
            raise NotImplementedError

        @property
        def stops(self):
            raise NotImplementedError

        @property
        def arrivals(self):
            raise NotImplementedError

    TestProvider(None)


def test_default_gtfs_db_path():
    c = Config({'busbus_dir': '/tmp/busbus'})
    assert c['gtfs_db_path'] == '/tmp/busbus/gtfs.sqlite3'


@responses.activate
def test_already_imported(provider, gtfs_zip_data):
    assert len(list(provider.conn.cursor().execute(
        'select id from _feeds'))) == 1
    e = busbus.Engine({'gtfs_db_path': provider.conn})

    responses.add(responses.GET, SampleGTFSProvider.gtfs_url,
                  body=gtfs_zip_data, status=200,
                  content_type='application/zip')
    p = SampleGTFSProvider(e)

    assert provider.conn is p.conn
    assert provider.feed_id == p.feed_id
    assert len(list(provider.agencies)) == 1


@pytest.mark.parametrize('entity', (None, busbus.Stop, busbus.Arrival))
def test_provider_get_default(provider, entity):
    assert provider.get(entity, u'The weather in london',
                        'asdfjkl') == 'asdfjkl'


def test_sql_entity_mixin_build_select():
    class FakeEntity(SQLEntityMixin):
        __table__ = 'fake'
        __field_map__ = OrderedDict([
            ('lorem', 'ipsum'),
            ('foo', 'bar'),
        ])

    assert (FakeEntity._build_select([]) ==
            'select ipsum as lorem, bar as foo from fake')
    assert (FakeEntity._build_select(['_feed_url', 'fake_id']) ==
            ('select ipsum as lorem, bar as foo from fake '
             'where _feed_url=? and fake_id=?'))


def test_sql_entity_eq(provider):
    class FakeEntity(SQLEntityMixin):
        pass

    a1 = provider.get(busbus.Agency, u'DTA')
    assert a1 == a1
    assert a1 is a1
    assert a1 != 'the weather in london'
    assert a1 != FakeEntity()
    a2 = provider.get(busbus.Agency, u'DTA')
    assert a1 == a2
    assert a2 == a1
    assert a1 is not a2


entity_len_params = [
    ('agencies', 1),
    ('stops', 9),
    ('routes', 5),
    # keep in mind this is from 2007-06-03T06:45:00-07:00 to 09:45:00
    ('arrivals', 139),
    ('alerts', 0),
]


@pytest.mark.parametrize('attr,count', entity_len_params)
def test_entity_len_engine(engine, provider, attr, count):
    with mock.patch.object(engine, '_providers', {'hi': provider}) as m:
        assert len(list(getattr(engine, attr))) == count


@pytest.mark.parametrize('attr,count', entity_len_params)
def test_entity_len_provider(provider, attr, count):
    assert len(list(getattr(provider, attr))) == count


def test_stops_etc(provider):
    stop = next(provider.stops)
    # Empty CSV fields should be coerced to None
    assert stop.description is None
    # stops.txt inherits agency timezone if blank
    assert stop.timezone == 'America/Los_Angeles'


def test_stops_no_children(provider):
    # none of the stops have children
    for stop in provider.stops:
        assert len(list(stop.children)) == 0


def test_stops_no_children_base(provider):
    stop = next(provider.stops)
    stop = busbus.Stop(**dict(stop))
    assert len(list(stop.children)) == 0


def test_agencies_unicode(provider):
    """
    Our CSV parser should be reading everything in as Unicode.
    """
    for agency in provider.agencies:
        for attr in agency.__attrs__:
            if getattr(agency, attr):
                assert isinstance(getattr(agency, attr), six.text_type)


def test_stops_latlon(provider):
    for stop in provider.stops:
        assert isinstance(stop.latitude, float)
        assert isinstance(stop.longitude, float)


def test_stop_distance_to_same(provider):
    stop = provider.get(busbus.Stop, u'AMV')
    assert stop.distance_to(stop) < 0.000005


def test_stop_distance_to_other(provider):
    stop1 = provider.get(busbus.Stop, u'AMV')
    stop2 = provider.get(busbus.Stop, u'STAGECOACH')
    assert 43.6 < stop1.distance_to(stop2) / 1000 < 43.7


def test_stop_distance_to_latlon(provider):
    stop = provider.get(busbus.Stop, u'AMV')
    assert 43.6 < stop.distance_to((36.915682, -116.751677)) / 1000 < 43.7
    assert 43.6 < stop.distance_to(36.915682, -116.751677) / 1000 < 43.7


def test_stop_routes(provider):
    routes = list(provider.get(busbus.Stop, u'AMV').routes)
    assert len(routes) == 1
    assert routes[0].id == u'AAMV'


def test_routes_agency(provider):
    for route in provider.routes:
        assert route.agency.id == 'DTA'


def test_route_stops(provider):
    stops = list(provider.get(busbus.Route, u'CITY').stops)
    assert len(stops) == 5
    assert set(stop.id for stop in stops) == set((u'STAGECOACH', u'NANAA',
                                                  u'NADAV', u'DADAN', u'EMSI'))


@pytest.mark.parametrize('time,stop_id,count', [
    # for STAGECOACH, 06:45-09:45:
    # STBA: 6 arrivals (every half hour)
    # CITY1: 2 arrivals from 06:45-07:59 (every half hour)
    #       10 arrivals from 08:00-09:45 (every 10 minutes)
    # CITY2: 3 arrivals from 06:45-07:59 (trip start time)
    #        9 arrivals from 08:00-09:45 (trip start time)
    # Sunday
    ('2007-06-03T06:45:00-07:00', u'STAGECOACH', 30),
    ('2007-06-03T06:45:00-07:00', u'AMV', 1),
    # Monday (FULLW exception)
    ('2007-06-04T06:45:00-07:00', u'STAGECOACH', 0),
    ('2007-06-04T06:45:00-07:00', u'AMV', 0),
    # Tuesday
    ('2007-06-05T06:45:00-07:00', u'STAGECOACH', 30),
    ('2007-06-05T06:45:00-07:00', u'AMV', 0)
])
def test_valid_arrivals(provider, time, stop_id, count):
    time = arrow.get(time)
    stop = provider.get(busbus.Stop, stop_id)
    assert (len(list(provider.arrivals.where(stop=stop, start_time=time))) ==
            count)


@pytest.mark.parametrize('time,count', [
    (arrow.get('2007-06-03T06:45:00-07:00'), 6),
    # becomes 2007-06-02T17:00:00-07:00
    (datetime.date(2007, 6, 3), 7),
    # becomes 2007-06-03T06:45:00-07:00
    (datetime.datetime(2007, 6, 3, 13, 45), 6),
])
def test_arrivals_weird_kwargs(provider, time, count):
    stop = provider.get(busbus.Stop, u'STAGECOACH')
    route = provider.get(busbus.Route, u'STBA')
    assert (len(list(provider.arrivals.where(stop=stop, route=route,
                                             start_time=time))) == count)
    # test dotted kwargs
    assert (len(list(provider.arrivals.where(
        **{'stop.id': u'STAGECOACH', 'route.id': u'STBA', 'start_time': time}
    ))) == count)


def test_arrivals_end_before_start(provider):
    assert (len(list(provider.arrivals.where(
        start_time=arrow.now(),
        end_time=arrow.now().replace(hours=-3)))) == 0)


def test_no_realtime(provider):
    with pytest.raises(StopIteration):
        next(provider.arrivals.where(realtime=True))



from busbus.util import RateLimitRequests

from datetime import datetime, timedelta
import httpbin
import pytest
import requests
import time
from wsgi_intercept import requests_intercept, add_wsgi_intercept


@pytest.fixture(scope='module')
def url_prefix(request):
    host = 'httpbin.org'
    port = 80
    requests_intercept.install()
    add_wsgi_intercept(host, port, lambda: httpbin.app)

    @request.addfinalizer
    def fin():
        requests_intercept.uninstall()

    return 'http://{0}:{1}/'.format(host, port)


def test_request(url_prefix):
    resp = requests.get(url_prefix + '/get')
    assert len(resp.json()['args']) == 0


def test_repeat_get_no_rate_limit(url_prefix):
    session = RateLimitRequests()
    resp1 = session.get(url_prefix + '/cookies')
    session.get(url_prefix + '/cookies/set?a=b')
    resp2 = session.get(url_prefix + '/cookies')
    assert resp1.json()['cookies'] != resp2.json()['cookies']


def test_get_interval(url_prefix):
    session = RateLimitRequests(interval=0.2)
    resp1 = session.get(url_prefix + '/get')
    t = datetime.now()
    resp2 = session.get(url_prefix + '/ip')
    assert datetime.now() - t > timedelta(seconds=0.1)


def test_repeat_get_url_interval(url_prefix):
    session = RateLimitRequests(url_interval=1)
    t = datetime.now()
    resp1 = session.get(url_prefix + '/cookies')
    session.get(url_prefix + '/cookies/set?a=b')
    resp2 = session.get(url_prefix + '/cookies')
    assert datetime.now() - t < timedelta(seconds=1)
    assert resp1 == resp2
    assert resp1.json()['cookies'] == resp2.json()['cookies']
    time.sleep(1)
    resp3 = session.get(url_prefix + '/cookies')
    assert resp1 != resp3
    assert resp1.json()['cookies'] != resp3.json()['cookies']


def test_post(url_prefix):
    session = RateLimitRequests()
    assert not session.post(url_prefix + '/post').json()['form']



from busbus.provider import ProviderBase
from busbus.provider.gtfs import GTFSMixin


class CTranProvider(GTFSMixin, ProviderBase):
    credit = 'C-TRAN'
    credit_url = 'http://www.c-tran.org/'
    country = 'US'

    gtfs_url = "http://www.c-tran.com/images/Google/GoogleTransitUpload.zip"

    def __init__(self, engine=None):
        super(CTranProvider, self).__init__(engine, self.gtfs_url)



import busbus
from busbus.provider.mbta import MBTAProvider
from .conftest import mock_gtfs_zip

import arrow
import os
from pkg_resources import resource_string
import pytest
import responses
from six.moves import urllib


@pytest.fixture(scope='module')
@responses.activate
def mbta_provider(engine):
    responses.add(responses.GET, MBTAProvider.gtfs_url,
                  body=mock_gtfs_zip('mbta'), status=200,
                  content_type='application/zip')
    return MBTAProvider('fake API key', engine)


# test that we are indeed using our local abridged copy of the GTFS feed
def test_len_routes(mbta_provider):
    assert len(list(mbta_provider.routes)) == 5


def test_route_with_only_shortname(mbta_provider):
    route = mbta_provider.get(busbus.Route, id=u'39')
    assert route.name == u'39'
    assert route.short_name is None


def test_route_with_shortname_and_name(mbta_provider):
    route = mbta_provider.get(busbus.Route, id=u'Logan-22')
    assert route.short_name == 'Shuttle'
    assert route.name == 'Massport Subway Shuttle (22)'


def test_agency_phoneword_e164(mbta_provider):
    agency = mbta_provider.get(busbus.Agency, id=u'2')
    assert agency.phone_e164 == '+18002356426'


def test_bu_central_children(mbta_provider):
    children = list(mbta_provider.get(busbus.Stop, u'place-bucen').children)
    ids = set(c.id for c in children)
    assert ids == set((u'70144', u'70145'))
    for child in children:
        assert child.parent.id == u'place-bucen'


def test_green_to_bu_gtfs(mbta_provider):
    stop = mbta_provider.get(busbus.Stop, u'place-bucen')
    route = mbta_provider.get(busbus.Route, u'Green-B')
    assert len(list(mbta_provider.arrivals.where(
        stop=stop, route=route, realtime=False,
        start_time=arrow.get('2015-03-10T14:00:00-04:00'),
        end_time=arrow.get('2015-03-10T16:00:00-04:00')))) == 32


def test_realtime_too_many_requests(mbta_provider):
    with pytest.raises(ValueError):
        next(mbta_provider.arrivals)


def realtime_response_arrs(p, start, stop_id=None, route_id=None):
    status = 200
    kwargs = {'start_time': arrow.get(start)}

    if stop_id is None and route_id is None:
        raise ValueError()
    elif route_id is None:
        endpoint = 'predictionsbystop'
        filename = '{0}_{1}_{2}.json'.format(endpoint, stop_id, start)
    else:
        endpoint = 'predictionsbyroute'
        filename = '{0}_{1}_{2}.json'.format(endpoint, route_id, start)

    if filename == ('predictionsbyroute_Logan-22_'
                    '2015-05-01T17:12:21-05:00.json'):
        status = 404

    if stop_id is not None:
        kwargs['stop'] = p.get(busbus.Stop, stop_id)
    if route_id is not None:
        kwargs['route'] = p.get(busbus.Route, route_id)

    url = p.mbta_realtime_url + endpoint
    data = resource_string(__name__, os.path.join('data', 'mbta', filename))

    responses.add(responses.GET, url, status=status, body=data,
                  content_type='application/json; charset=utf-8')
    return p.arrivals.where(**kwargs)


@responses.activate
def test_realtime_39(mbta_provider):
    arrs = list(realtime_response_arrs(
        mbta_provider, '2015-05-01T15:10:23-05:00', route_id='39'))
    assert arrs[0].realtime is True
    assert arrs[0].time == arrow.get(1430511060)
    assert len(set(arr.stop.id for arr in arrs)) > 1


@responses.activate
def test_realtime_39_forest_hills(mbta_provider):
    arrs = list(realtime_response_arrs(
        mbta_provider, '2015-05-01T15:10:23-05:00',
        route_id='39', stop_id='8750'))
    assert len(arrs) == 57
    assert arrs[0].realtime is True
    assert arrs[0].time == arrow.get(1430511060)
    # MBTA doesn't provide realtime data for ~3 hours into the future
    assert arrs[-1].realtime is False


@responses.activate
def test_realtime_downtown_crossing(mbta_provider):
    arrs = list(realtime_response_arrs(
        mbta_provider, '2015-05-02T00:04:05-05:00', stop_id='place-dwnxg'))
    # this query was performed late at night when all remaining arrivals should
    # have realtime data (5 orange line, 5 red line arrivals returned)
    assert len(arrs) == 5
    assert all(arr.realtime is True for arr in arrs)
    # although the response contains red line arrivals, we removed the red line
    # from our abridged version of the feed
    assert all(arr.route.id == 'Orange' for arr in arrs)


@responses.activate
def test_realtime_logan_22(mbta_provider):
    arrs = list(realtime_response_arrs(
        mbta_provider, '2015-05-01T17:12:21-05:00', route_id='Logan-22'))
    assert len(arrs) == 120
    assert all(arr.realtime is False for arr in arrs)


@responses.activate
def test_predictionsbyroute_children(mbta_provider):
    arrs = list(realtime_response_arrs(
        mbta_provider, '2015-05-01T15:08:18-05:00',
        stop_id='place-dwnxg', route_id='Orange'))
    assert len(arrs) > 0



import busbus
from busbus.entity import BaseEntityJSONEncoder

import json
import pytest


@pytest.fixture(scope='module')
def agency(provider):
    return next(provider.agencies)


def test_entity_repr(agency):
    assert 'DTA' in repr(agency)


def test_entity_failed_getattr(agency):
    with pytest.raises(AttributeError):
        agency.the_weather_in_london


def test_entity_failed_getitem(agency):
    with pytest.raises(KeyError):
        agency['the_weather_in_london']


def test_provider_failed_getattr(provider):
    with pytest.raises(AttributeError):
        provider.the_weather_in_london


def test_provider_failed_getitem(provider):
    with pytest.raises(KeyError):
        provider['the_weather_in_london']


def test_entity_to_dict(agency):
    assert dict(agency)['id'] == 'DTA'


def test_entity_to_json(provider):
    json_str = BaseEntityJSONEncoder().encode(next(provider.arrivals))
    json.loads(json_str)


def test_bad_json():
    with pytest.raises(TypeError):
        BaseEntityJSONEncoder().encode(busbus.Engine)



from busbus.entity import BaseEntity
from busbus.queryable import Queryable
from busbus.util import Config, dist

import errno
import os
import six


class Engine(object):
    """
    The busbus.Engine manages data providers and allows querying entities and
    subscribing to entity events the providers.
    """

    def __init__(self, config=None):
        self.config = Config(config)
        self._providers = {}

        try:
            os.mkdir(self.config['busbus_dir'])
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

    def _register_provider(self, provider):
        self._providers[provider.id] = provider

    @property
    def providers(self):
        return Queryable(self._providers.values())

    @property
    def agencies(self):
        return Queryable.chain(*[p.agencies for p in self._providers.values()])

    @property
    def stops(self):
        return Queryable.chain(*[p.stops for p in self._providers.values()])

    @property
    def routes(self):
        return Queryable.chain(*[p.routes for p in self._providers.values()])

    @property
    def arrivals(self):
        return Queryable.chain(*[p.arrivals for p in self._providers.values()])

    @property
    def alerts(self):
        return Queryable.chain(*[p.alerts for p in self._providers.values()])


class Agency(BaseEntity):
    __attrs__ = ('id', 'name', 'url', 'timezone', 'lang', 'phone_e164',
                 'phone_human', 'fare_url')


class Stop(BaseEntity):
    __attrs__ = ('id', 'code', 'name', 'description', 'latitude', 'longitude',
                 'zone', 'url', 'parent', 'timezone', 'accessible')

    @property
    def children(self):
        return self.provider.stops.where(lambda s: s.parent == self)

    @staticmethod
    def add_children(it):
        """
        Given an iterable of stops, yields the stops including all their
        children.
        """
        yielded_ids = []
        stops = list(it)
        while stops:
            stop = stops.pop(0)
            if stop.id in yielded_ids:
                continue
            yielded_ids.append(stop.id)
            yield stop
            stops.extend(stop.children)

    def distance_to(self, *args):
        """
        Returns the distance to another stop (one Stop argument) or a latitude
        and longitude (two arguments, or one two-element argument) in meters.
        """
        if len(args) == 1:
            if isinstance(args[0], Stop):
                return dist(self.latitude, self.longitude,
                            args[0].latitude, args[0].longitude)
            else:
                return dist(self.latitude, self.longitude, *args[0])
        else:
            return dist(self.latitude, self.longitude, *args)


class Route(BaseEntity):
    __attrs__ = ('id', 'agency', 'short_name', 'name', 'description', 'type',
                 'url', 'color', 'text_color')


class Arrival(BaseEntity):
    __attrs__ = ('route', 'stop', 'time', 'departure_time', 'headsign',
                 'short_name', 'bikes_ok', 'realtime')
    __repr_attrs__ = ('route', 'stop', 'time')

    def __lt__(self, other):
        return self.time < other.time


class Alert(BaseEntity):
    __attrs__ = ('id', 'text')


ENTITIES = (Agency, Stop, Route, Arrival, Alert)



import busbus
from busbus.provider import ProviderBase
from busbus.provider.gtfs import GTFSMixin, GTFSArrivalGenerator
from busbus.util import RateLimitRequests
from busbus.util.arrivals import ArrivalQueryable, ArrivalGeneratorBase

import arrow
import collections
import heapq
import itertools
import six


class MBTAArrivalGenerator(ArrivalGeneratorBase):
    realtime = True

    def _build_iterable(self):
        self.gtfs_gen = GTFSArrivalGenerator(self.provider, self.stops,
                                             self.routes, self.start, self.end)

        if self.stops is None and self.routes is None:
            raise ValueError('Provide at least one stop or route for realtime '
                             'data, or set realtime=False')
        elif self.routes is None:
            def yield_predictions_by_stop(stop):
                cur = self.provider.conn.cursor()
                routes = stop.routes

                resp = self.provider._mbta_realtime_call('predictionsbystop',
                                                         {'stop': stop.id})
                if resp.status_code == 200:
                    trips = {}
                    for mode in resp.json()['mode']:
                        for route in mode['route']:
                            for direction in route['direction']:
                                for trip in direction['trip']:
                                    query = cur.execute(
                                        '''select null from stop_times where
                                        stop_id=? and trip_id=? and _feed=?''',
                                        (stop.id, trip['trip_id'],
                                         self.provider.feed_id))
                                    # if the query returns anything it's
                                    # guaranteed to be 1 row
                                    for row in query:
                                        trips[trip['trip_id']] = {
                                            'pre_dt': trip['pre_dt'],
                                            'headsign': trip['trip_headsign'],
                                            'route_id': route['route_id'],
                                            'stop_id': stop.id,
                                        }
                    its = [self._merge_arrivals(stop, route, trips)
                           for route in routes]
                else:
                    its = [self._merge_arrivals(stop, route, None)
                           for route in routes]
                return heapq.merge(*its)

            its = list(six.moves.map(yield_predictions_by_stop,
                                     busbus.Stop.add_children(self.stops)))
        else:
            def yield_predictions_by_route(route):
                if self.stops is None:
                    stops = route.stops
                else:
                    stops = self.stops
                stops = list(busbus.Stop.add_children(stops))

                resp = self.provider._mbta_realtime_call('predictionsbyroute',
                                                         {'route': route.id})
                if resp.status_code == 200:
                    trips = {stop.id: dict() for stop in stops}
                    for dir in resp.json()['direction']:
                        for trip in dir['trip']:
                            for stop in trip['stop']:
                                if (stop['stop_sequence'] != '0' and
                                        stop['stop_id'] in trips):
                                    trips[stop['stop_id']][trip['trip_id']] = {
                                        'pre_dt': stop['pre_dt'],
                                        'headsign': trip['trip_headsign'],
                                        'route_id': route.id,
                                        'stop_id': stop['stop_id'],
                                    }
                    its = [self._merge_arrivals(stop, route, trips[stop.id])
                           for stop in stops if trips[stop.id]]
                else:
                    its = [self._merge_arrivals(stop, route, None)
                           for stop in stops]
                return heapq.merge(*its)

            its = list(six.moves.map(yield_predictions_by_route, self.routes))
        return heapq.merge(*its)

    def _merge_arrivals(self, stop, route, trips):
        arrs = collections.defaultdict(list)
        for tid, arr in self._build_scheduled_arrivals(stop, route):
            arrs[tid].append(arr)
        if trips is not None:
            for tid, arr in self._build_realtime_arrivals(stop, route, trips):
                arrs[tid] = [arr]
        arrs = itertools.chain(*arrs.values())
        return sorted(arrs)

    def _build_realtime_arrivals(self, stop, route, trips):
        for trip_id, trip in trips.items():
            if stop.id == trip['stop_id'] and route.id == trip['route_id']:
                time = arrow.get(trip['pre_dt']).to(self.provider._timezone)
                if self.start <= time <= self.end:
                    arr = busbus.Arrival(self.provider, realtime=True,
                                         stop=stop, route=route,
                                         time=time, departure_time=time,
                                         headsign=trip['headsign'])
                    yield (trip_id, arr)

    def _build_scheduled_arrivals(self, stop, route):
        for stop_time in self.gtfs_gen._stop_times(stop, route):
            for arr in self.gtfs_gen._build_arrivals(stop, route, stop_time):
                if self.start <= arr.time <= self.end:
                    yield (stop_time['trip_id'], arr)


class MBTAProvider(GTFSMixin, ProviderBase):
    """
    Provides data from the MBTA-realtime API v2 and the MBTA GTFS feed.

    http://realtime.mbta.com/

    An API key is required to access the MBTA-realtime API. MBTA provides a key
    you can use for development at:
        http://realtime.mbta.com/Portal/Content/Download/APIKey.txt
    This key is for development only -- MBTA requests you do not use this key
    in production; request a key instead.
    """

    legal = [
        ('http://www.massdot.state.ma.us/Portals/0/docs/developers/'
         'RelationshipPrinciples.pdf'),
        ('http://www.massdot.state.ma.us/Portals/0/docs/developers/'
         'develop_license_agree.pdf'),
    ]
    credit = 'MassDOT'
    credit_url = 'http://www.massdot.state.ma.us'
    country = 'US'

    gtfs_url = "http://www.mbta.com/uploadedfiles/MBTA_GTFS.zip"
    mbta_realtime_url = "http://realtime.mbta.com/developer/api/v2/"

    def __init__(self, mbta_api_key, engine=None):
        super(MBTAProvider, self).__init__(engine, self.gtfs_url)
        self.mbta_api_key = mbta_api_key

        # MBTA requires that "the same polling command" is not to be called
        # more often than every 10 seconds (API docs, "Use of MBTA data").
        self._requests = RateLimitRequests(url_interval=10)

    def _mbta_realtime_call(self, query, params):
        url = self.mbta_realtime_url + query
        params.update({'api_key': self.mbta_api_key, 'format': 'json'})
        return self._requests.get(url, params=params)

    @property
    def arrivals(self):
        return ArrivalQueryable(self, (MBTAArrivalGenerator,
                                       GTFSArrivalGenerator))



import busbus
from busbus.provider.ctran import CTranProvider
from .conftest import mock_gtfs_zip

import arrow
import pytest
import responses


@pytest.fixture(scope='module')
@responses.activate
def ctran_provider(engine):
    responses.add(responses.GET, CTranProvider.gtfs_url,
                  body=mock_gtfs_zip('ctran'), status=200,
                  content_type='application/zip')
    return CTranProvider(engine)


# test that we are indeed using our local abridged copy of the GTFS feed
def test_len_routes(ctran_provider):
    assert len(list(ctran_provider.routes)) == 28


@pytest.mark.parametrize('stop_id,count', [
    (u'2058', 4)
])
def test_43_to_eaton_hall(ctran_provider, stop_id, count):
    stop = ctran_provider.get(busbus.Stop, stop_id)
    route = ctran_provider.get(busbus.Route, u'46')
    assert len(list(ctran_provider.arrivals.where(
        stop=stop, route=route,
        start_time=arrow.get('2015-03-10T14:00:00-05:00'),
        end_time=arrow.get('2015-03-10T16:00:00-05:00')))) == count



import busbus
import busbus.entity
from busbus import util

import pytest


def test_util_clsname():
    engine = busbus.Engine()
    assert util.clsname(engine) == 'busbus.Engine'


def test_util_entity_type_error():
    engine = busbus.Engine()
    with pytest.raises(TypeError):
        busbus.util.entity_type(busbus.entity.BaseEntity)
    with pytest.raises(TypeError):
        busbus.util.entity_type(engine)



import busbus
from busbus.util import clsname

from abc import ABCMeta, abstractmethod, abstractproperty
from cachecontrol import CacheControl
from cachecontrol.caches import FileCache
import hashlib
import requests
import six
import uuid


@six.add_metaclass(ABCMeta)
class ProviderBase(object):

    def __init__(self, engine, **kwargs):
        self.id = hashlib.sha1(six.b(
            '{0}:{1!r}'.format(clsname(self), kwargs.get('__init_args__', {}))
        )).hexdigest()

        if engine:
            self.engine = engine
        else:
            self.engine = busbus.Engine()
        self.engine._register_provider(self)

        self._requests = requests.Session()

        # This requests session object, wrapped with CacheControl, is useful
        # for long-term storage of larger files, such as GTFS data.
        self._cached_requests = CacheControl(self._requests, cache=FileCache(
            self.engine.config['url_cache_dir']))

    def _new_entity(self, entity):
        """
        Method to receive knowledge of a new entity -- does nothing if not
        overridden
        """

    @abstractmethod
    def get(self, entity, id, default=None):
        """Return the requested entity, or default if it doesn't exist"""

    @abstractproperty
    def agencies(self):
        """Return an iterator of the agencies for this provider"""

    @abstractproperty
    def stops(self):
        """Return an iterator of the stops for this provider"""

    @abstractproperty
    def routes(self):
        """Return an iterator of the routes for this provider"""

    @abstractproperty
    def arrivals(self):
        """Return an iterator of the arrivals for this provider"""

    @property
    def alerts(self):
        """Return an iterator of current alerts for this provider"""
        return iter(())

    def __getitem__(self, name):
        try:
            return getattr(self, name)
        except AttributeError:
            raise KeyError(name)

    def keys(self):
        for attr in ('id', 'legal', 'credit', 'credit_url', 'country'):
            if getattr(self, attr, None) is not None:
                yield attr



import busbus

import os
import pytest
import tempfile


def test_engine_busbus_dir_exists():
    dir = tempfile.mkdtemp()
    engine = busbus.Engine({'busbus_dir': dir})
    os.rmdir(dir)


def test_engine_busbus_dir_other_error():
    outerdir = tempfile.mkdtemp()
    dir = os.path.join(outerdir, 'inner')
    os.mkdir(dir)
    old_mode = os.stat(outerdir).st_mode
    os.chmod(outerdir, int('400', 8))
    with pytest.raises(OSError):
        engine = busbus.Engine({'busbus_dir': dir})
    os.chmod(outerdir, old_mode)
    os.rmdir(dir)
    os.rmdir(outerdir)



from busbus.provider import ProviderBase
from busbus.provider.gtfs import GTFSMixin


class LawrenceTransitProvider(GTFSMixin, ProviderBase):
    credit = 'Lawrence Transit'
    credit_url = 'http://www.lawrencetransit.org/'
    country = 'US'

    gtfs_url = ("http://lawrenceks.org/assets/gis/google-transit/"
                "google_transit.zip")

    def __init__(self, engine=None):
        super(LawrenceTransitProvider, self).__init__(engine, self.gtfs_url)



from __future__ import division

import busbus
import busbus.entity
from busbus.queryable import Queryable
from busbus import util
from busbus.util.arrivals import ArrivalQueryable, ArrivalGeneratorBase
from busbus.util.csv import CSVReader

import apsw
import arrow
import collections
import datetime
import hashlib
import heapq
import itertools
import operator
import os
import phonenumbers
from pkg_resources import resource_string
import six
import zipfile


# This must be the same as the user_version pragma in gtfs.sql
SCHEMA_USER_VERSION = 2015020201


def parse_gtfs_time(timestr):
    """
    GTFS time strings are HH:MM:SS (or H:MM:SS if the hours field is less than
    10). The time is relative to noon on a given day, and the hours field can
    go over 23 to represent a time on the following day (GTFS time never
    decreases in a given trip).

    This function is not quite that picky, but its behavior is undefined if
    it's given a string not following the specification.

    This function returns a timedelta, which contains the number of seconds
    since noon represented by a time on a given day.
    """
    split = timestr.split(':')
    seconds = int(split[-1])
    minutes = int(split[-2]) if len(split) > 1 else 0
    hours = int(split[-3]) if len(split) > 2 else 0
    return (hours - 12) * 3600 + minutes * 60 + seconds


def date_to_sql_string(datestr):
    return '-'.join((datestr[:4], datestr[4:6], datestr[6:]))


FIX_TYPE_MAP = {
    'date': date_to_sql_string,
    'gtfstime': parse_gtfs_time,
    'integer': int,
    'real': float,
    'timedelta': int,
    'text': lambda s: s,
}


class SQLEntityMixin(object):

    def __eq__(self, other):
        if not isinstance(other, SQLEntityMixin):
            return False
        return (self.id == getattr(other, 'id', not self.id) and
                self.provider == getattr(other, 'provider', not self.provider))

    @classmethod
    def _build_select(cls, columns, named_params=False):
        query = 'select {0} from {1}'.format(
            ', '.join('{1} as {0}'.format(k, v)
                      for k, v in cls.__field_map__.items()),
            cls.__table__)
        if columns:
            if named_params:
                query += ' where {0}'.format(
                    ' and '.join('{0}=:{0}'.format(c) for c in columns))
            else:
                query += ' where {0}'.format(
                    ' and '.join('{0}=?'.format(c) for c in columns))
        return query

    def _query(self, **kwargs):
        return self.provider._query(self.__class__, **kwargs)

    @classmethod
    def from_id(cls, provider, id, default=None):
        """
        Special case of _query that doesn't require an instantiated object
        and to fetch by a specific column name.
        """
        id_field = cls.__field_map__['id']
        result = provider._query(cls, **{id_field: id}).fetchone()
        if result is None:
            return default
        else:
            return cls(provider, **dict(result))


class GTFSAgency(SQLEntityMixin, busbus.Agency):
    __table__ = 'agency'
    __field_map__ = {
        'id': 'agency_id',
        'name': 'agency_name',
        'url': 'agency_url',
        'timezone': 'agency_timezone',
        'lang': 'agency_lang',
        'phone_human': 'agency_phone',
        'fare_url': 'agency_fare_url',
    }

    def __init__(self, provider, **data):
        if data.get('lang'):
            data['lang'] = data['lang'].lower()
        if data.get('phone_human') and getattr(provider, 'country'):
            phone = phonenumbers.parse(data['phone_human'], provider.country)
            data['phone_e164'] = phonenumbers.format_number(
                phone, phonenumbers.PhoneNumberFormat.E164)

        super(GTFSAgency, self).__init__(provider, **data)


class GTFSStop(SQLEntityMixin, busbus.Stop):
    __table__ = 'stops'
    __field_map__ = {
        'id': 'stop_id',
        'code': 'stop_code',
        'name': 'stop_name',
        'description': 'stop_desc',
        'latitude': 'stop_lat',
        'longitude': 'stop_lon',
        'url': 'stop_url',
        '_parent_id': 'parent_station',
        'timezone': 'stop_timezone',
        '_accessible': 'wheelchair_boarding',
    }

    def __init__(self, provider, **data):
        if data.get('_parent_id'):
            data['parent'] = busbus.entity.LazyEntityProperty(
                provider.get, busbus.Stop, data['_parent_id'])
        if not data.get('timezone'):
            data['timezone'] = provider._timezone
        if '_accessible' in data:
            pass  # FIXME

        super(GTFSStop, self).__init__(provider, **data)

    @property
    def routes(self):
        result = self.provider.conn.cursor().execute(
            '''select route_id from _stops_routes where
            stop_id=? and _feed=?''',
            (self.id, self.provider.feed_id))
        return Queryable(self.provider.get(busbus.Route, row['route_id'])
                         for row in result)

    @property
    def children(self):
        result = self._query(parent_station=self.id)
        return Queryable(GTFSStop(self.provider, **dict(row))
                         for row in result)


class GTFSRoute(SQLEntityMixin, busbus.Route):
    __table__ = 'routes'
    __field_map__ = {
        'id': 'route_id',
        '_agency_id': 'agency_id',
        'short_name': 'route_short_name',
        'name': 'route_long_name',
        'description': 'route_desc',
        '_type': 'route_type',
        'url': 'route_url',
        'color': 'route_color',
        'text_color': 'route_text_color',
    }

    def __init__(self, provider, **data):
        if '_agency_id' in data:
            data['agency'] = busbus.entity.LazyEntityProperty(
                provider.get, busbus.Agency, data['_agency_id'])
        if '_type' in data:
            pass  # FIXME
        if data.get('name') is None:
            data['name'] = data.pop('short_name')

        super(GTFSRoute, self).__init__(provider, **data)

    @property
    def stops(self):
        result = self.provider.conn.cursor().execute(
            '''select stop_id from _stops_routes where
            route_id=? and _feed=?''',
            (self.id, self.provider.feed_id))
        return Queryable(self.provider.get(busbus.Stop, row['stop_id'])
                         for row in result)

    @property
    def directions(self):
        hashes = []
        t_query = """select trip_headsign, trip_short_name, bikes_allowed,
        trip_id from trips where route_id=:route_id and _feed=:_feed"""
        t_filter = {'route_id': self.id, '_feed': self.provider.feed_id}
        cur = self.provider.conn.cursor()
        for trip in cur.execute(t_query, t_filter):
            direction = {}
            innercur = self.provider.conn.cursor()
            result = innercur.execute(
                """select s.* from stop_times as st join stops as s
                on st.stop_id=s.stop_id and st._feed=s._feed
                where st.trip_id=:t_id and st._feed=:_feed""",
                {'t_id': trip['trip_id'], '_feed': self.provider.feed_id})
            direction['stops'] = [GTFSStop(self.provider, **dict(row))
                                  for row in result]
            if trip['trip_headsign'] is not None:
                direction['headsign'] = trip['trip_headsign']
            if trip['trip_short_name'] is not None:
                direction['short_name'] = trip['trip_short_name']
            if trip['bikes_allowed'] is not None:
                direction['bikes_ok'] = trip['bikes_allowed']
            h = util.freezehash(direction)
            if h not in hashes:
                hashes.append(h)
                yield direction


class GTFSArrivalGenerator(ArrivalGeneratorBase):
    realtime = False

    def __init__(self, provider, stops, routes, start, end):
        super(GTFSArrivalGenerator, self).__init__(provider, stops, routes,
                                                   start, end)

        if self.stops is None:
            if self.routes is None:
                self.stops = self.provider.stops
                self.routes = self.provider.routes
            else:
                stops_dict = {}
                for route in self.routes:
                    for stop in route.stops:
                        stops_dict[stop.id] = stop
                self.stops = stops_dict.values()
        else:
            if self.routes is None:
                routes_dict = {}
                for stop in self.stops:
                    for route in stop.routes:
                        routes_dict[route.id] = route
                self.routes = routes_dict.values()

        self.service_cache = {}
        self.freq_cache = {}
        self.it = None

    def _stop_times(self, stop, route):
        return self.provider.conn.cursor().execute(
            '''select t.*, arr, departure_time from
                (select trip_id, _min_arrival_time, service_id, trip_headsign,
                trip_short_name, bikes_allowed from trips where
                route_id=:route_id and _feed=:_feed) as t
            join
                (select trip_id, coalesce(arrival_time, _arrival_interpolate)
                as arr, departure_time from stop_times where stop_id=:stop_id
                and _feed=:_feed) as st
            on t.trip_id=st.trip_id order by arr asc''',
            {'stop_id': stop.id, 'route_id': route.id,
             '_feed': self.provider.feed_id})

    def _build_iterable(self):
        iters = []
        stops = busbus.Stop.add_children(self.stops)
        for stop, route in itertools.product(stops, self.routes):
            for stop_time in self._stop_times(stop, route):
                iters.append(self._build_arrivals(stop, route, stop_time))
        return heapq.merge(*iters)

    def _build_arrivals(self, stop, route, stop_time):
        def build_arr(day, offset=None):
            if offset is None:
                offset = datetime.timedelta()
            time = day + datetime.timedelta(seconds=stop_time['arr']) + offset
            if not (self.start <= time <= self.end):
                return
            dep = (day + stop_time['departure_time'] + offset if
                   stop_time['departure_time'] else None)
            bikes_ok = {1: True, 2: False}.get(stop_time['bikes_allowed'])
            return busbus.Arrival(self.provider, stop=stop, route=route,
                                  time=time, departure_time=dep,
                                  headsign=stop_time['trip_headsign'],
                                  short_name=stop_time['trip_short_name'],
                                  bikes_ok=bikes_ok, realtime=False)

        days = filter(self._valid_date_filter(stop_time['service_id']),
                      arrow.Arrow.range('day', self.start.floor('day'),
                                        self.end.ceil('day')))
        freqs = self._frequencies(stop_time['trip_id'])
        trip_start = stop_time['_min_arrival_time']
        for day in days:
            # GTFS time is relative to noon
            day = day.replace(hours=12)
            for freq in freqs:
                freq_start = day + freq['start_time']
                freq_end = day + freq['end_time']
                rel_time = freq_start - (day + trip_start)
                offset = datetime.timedelta()
                while freq_start + offset <= freq_end:
                    arrival = build_arr(day, offset + rel_time)
                    if arrival:
                        yield arrival
                    offset += freq['headway_secs']
            if not freqs:
                arrival = build_arr(day)
                if arrival:
                    yield arrival

    def _valid_date_filter(self, service_id):
        def valid_date(day):
            serv = self._service(service_id)
            weekday = day.format('dddd').lower()
            day = day.date()  # convert from Arrow to datetime.date
            # schedule exceptions: 1 = added, 2 = removed
            return ((serv['start_date'] <= day <= serv['end_date']) and
                    (serv['exceptions'].get(day, 0) != 2) and
                    (serv[weekday] or serv['exceptions'].get(day, 0) == 1))
        return valid_date

    def _frequencies(self, trip_id):
        if trip_id not in self.freq_cache:
            query = """select start_time, end_time, headway_secs
            from frequencies where trip_id=:trip_id and
            _feed=:_feed order by start_time asc"""
            filter = {'trip_id': trip_id, '_feed': self.provider.feed_id}
            cur = self.provider.conn.cursor()
            self.freq_cache[trip_id] = cur.execute(query, filter).fetchall()
        return self.freq_cache[trip_id]

    def _service(self, id):
        if id not in self.service_cache:
            c_query = """select start_date, end_date, monday, tuesday,
            wednesday, thursday, friday, saturday, sunday from calendar
            where service_id=:service_id and _feed=:_feed"""
            c_filter = {'service_id': id, '_feed': self.provider.feed_id}
            cur = self.provider.conn.cursor()
            self.service_cache[id] = dict(next(cur.execute(c_query, c_filter)))

            cd_query = """select date, exception_type as e from calendar_dates
            where service_id=:service_id and _feed=:_feed"""
            cd_result = cur.execute(cd_query, c_filter)
            self.service_cache[id]['exceptions'] = {r['date']: r['e']
                                                    for r in cd_result}
        return self.service_cache[id]


def gtfs_row_tracer(cur, row):
    type_map = {
        'date': lambda s: datetime.date(int(s[:4]), int(s[5:7]), int(s[8:])),
        'gtfstime': lambda i: datetime.timedelta(seconds=i),
        'timedelta': lambda i: datetime.timedelta(seconds=i),
    }
    desc = cur.getdescription()
    return {desc[i][0]: (type_map.get(desc[i][1], lambda s: s)(x)
                         if row[i] is not None else None)
            for i, x in enumerate(row)}


class GTFSMixin(object):
    """
    Mixin to parse transit data from a General Transit Feed Specification feed.

    GTFS is defined at https://developers.google.com/transit/gtfs/
    """

    def __init__(self, engine, gtfs_url):
        super(GTFSMixin, self).__init__(engine)

        if isinstance(self.engine.config['gtfs_db_path'], apsw.Connection):
            self.conn = self.engine.config['gtfs_db_path']
        else:
            self.conn = apsw.Connection(self.engine.config['gtfs_db_path'])
        self.conn.setrowtrace(gtfs_row_tracer)
        cur = self.conn.cursor()

        version = next(cur.execute('pragma user_version'))['user_version']
        if version == 0:
            script = resource_string(
                __name__, 'gtfs_{0}.sql'.format(SCHEMA_USER_VERSION))
            if isinstance(script, six.binary_type):
                script = script.decode('utf-8')
            cur.execute(script)
        elif version < SCHEMA_USER_VERSION:
            raise NotImplementedError()
        elif version > SCHEMA_USER_VERSION:
            raise RuntimeError('Database version is {0}, but only version {1} '
                               'is known'.format(version, SCHEMA_USER_VERSION))

        tables = [r['name'] for r in cur.execute('select name from '
                                                 'sqlite_master where '
                                                 'type="table"')
                  if not r['name'].startswith('_')]

        if isinstance(gtfs_url, six.binary_type):
            gtfs_url = gtfs_url.decode('utf-8')
        resp = self._cached_requests.get(gtfs_url)
        zip = six.BytesIO(resp.content).getvalue()
        hash = hashlib.sha256(zip).hexdigest()

        resp = [x['id'] for x in cur.execute(
            'select id from _feeds where url=? AND sha256sum=?',
            (gtfs_url, hash))]
        if len(resp) == 1:
            self.feed_id = resp[0]
        else:
            cur.execute('begin transaction')
            old_ids = [(x['id'],) for x in cur.execute(
                '''select id from _feeds where url=?''', (gtfs_url,))]
            cur.executemany('delete from _feeds where id=?', old_ids)
            for table in tables:
                cur.executemany('delete from {0} where _feed=?'.format(table),
                                old_ids)

            cur.execute('insert into _feeds (url, sha256sum) '
                        'values (?, ?)', (gtfs_url, hash))
            self.feed_id = self.conn.last_insert_rowid()
            with zipfile.ZipFile(six.BytesIO(zip)) as z:
                for table in tables:
                    filename = table + '.txt'
                    if filename not in z.namelist():
                        continue
                    with z.open(filename) as f:
                        data = CSVReader(f)
                        columns = []
                        coldata = []
                        for x in cur.execute(
                                'pragma table_info({0})'.format(table)):
                            if x['name'] in data.header:
                                columns.append(x['name'])
                                coldata.append((
                                    x['type'], data.header.index(x['name'])))
                        # _feed must be at end
                        columns.append('_feed')

                        stmt = ('insert into {0} ({1}) values ({2})'
                                .format(table, ', '.join(columns),
                                        ', '.join(('?',) * len(columns))))

                        def row_gen(row):
                            for t, idx in coldata:
                                if idx < len(row) and row[idx] is not None:
                                    yield FIX_TYPE_MAP[t](row[idx])
                                else:
                                    yield None
                            yield self.feed_id
                        cur.executemany(stmt, (row_gen(row) for row in data))
            cur.execute('commit transaction')

            cur.execute('begin transaction')
            for row in cur.execute(
                    '''select trip_id from trips where _feed=?''',
                    (self.feed_id,)):
                innercur = self.conn.cursor()
                trip_id = row['trip_id']

                # interpolate missing stop times
                known_times = {r['seq']: dict(r) for r in innercur.execute(
                    '''select arrival_time as a, departure_time as d,
                    stop_sequence as seq from stop_times where trip_id=:trip_id
                    and _feed=:_feed and arrival_time is not null
                    order by stop_sequence asc''',
                    {'trip_id': trip_id, '_feed': self.feed_id})}
                if not known_times:
                    # this trip is headway only
                    continue
                unknown_times = [r['stop_sequence'] for r in innercur.execute(
                    '''select stop_sequence from stop_times where
                    trip_id=:trip_id and _feed=:_feed and
                    arrival_time is null order by stop_sequence asc''',
                    {'trip_id': trip_id, '_feed': self.feed_id})]
                for i, seq in enumerate(unknown_times):
                    left = max(filter(lambda k: k < seq, known_times))
                    right = min(filter(lambda k: k > seq, known_times))
                    start = known_times[left].get('d', known_times[left]['a'])
                    gap = known_times[right]['a'] - start
                    count = len(list(filter(lambda k: left < k < right,
                                            unknown_times))) + 1
                    time = ((gap.total_seconds() * (i + 1) / count) +
                            start.total_seconds())
                    innercur.execute(
                        '''update stop_times set _arrival_interpolate=:a where
                        trip_id=:trip_id and _feed=:_feed and
                        stop_sequence=:seq''',
                        {'trip_id': trip_id, '_feed': self.feed_id,
                         'a': time, 'seq': seq})

                # fill in _min_arrival_time
                min_time = innercur.execute(
                    '''select min(coalesce(arrival_time, _arrival_interpolate))
                    as min_time from stop_times where trip_id=? and _feed=?''',
                    (trip_id, self.feed_id)).fetchone()['min_time']
                innercur.execute('''update trips set _min_arrival_time=?
                                 where trip_id=? and _feed=?''',
                                 (min_time, trip_id, self.feed_id))

            cur.execute(
                '''insert into _stops_routes (stop_id, route_id, _feed)
                select distinct st.stop_id, t.route_id, t._feed from
                (select trip_id, stop_id from stop_times where _feed=:_feed)
                as st join
                (select trip_id, route_id, _feed from trips where _feed=:_feed)
                as t on st.trip_id=t.trip_id''', {'_feed': self.feed_id})
            cur.execute('commit transaction')

    def _query(self, cls, **kwargs):
        if '_feed' not in kwargs:
            kwargs['_feed'] = self.feed_id
        cur = self.conn.cursor()
        return cur.execute(cls._build_select(
            kwargs.keys(), named_params=True), kwargs)

    def _entity_builder(self, cls, **kwargs):
        query = self._query(cls, **kwargs)
        return Queryable(cls(self, **row) for row in query)

    def get(self, cls, id, default=None):
        typemap = {
            busbus.Agency: GTFSAgency,
            busbus.Stop: GTFSStop,
            busbus.Route: GTFSRoute,
        }
        try:
            cls = util.entity_type(cls)
        except TypeError:
            return default
        if cls in typemap:
            return typemap[cls].from_id(self, id, default)
        else:
            return default

    @property
    def _timezone(self):
        for agency in self.agencies:
            if agency.timezone:
                return agency.timezone
        return None

    @property
    def agencies(self):
        return self._entity_builder(GTFSAgency)

    @property
    def stops(self):
        return self._entity_builder(GTFSStop)

    @property
    def routes(self):
        return self._entity_builder(GTFSRoute)

    @property
    def arrivals(self):
        return ArrivalQueryable(self, GTFSArrivalGenerator)



from busbus import util

from copy import copy
import itertools


def check_obj_attrs(obj, mapping):
    for k, v in mapping.items():
        if hasattr(obj, k):
            if getattr(obj, k) != v:
                return False
        elif (hasattr(obj, '__getitem__') and
              hasattr(obj, '__contains__')):
            if not (k in obj and obj[k] == v):
                return False
        else:
            return False
    return True


class Queryable(util.Iterable):

    def __init__(self, it, query_funcs=None, **kwargs):
        self.it = iter(it)
        self.query_funcs = tuple(query_funcs) if query_funcs else ()
        self.kwargs = kwargs

    def __next__(self):
        while True:
            value = next(self.it)
            if (check_obj_attrs(value, self.kwargs) and
                    all(query_func(value) for query_func in self.query_funcs)):
                return value

    def _new(self, query_funcs, kwargs):
        return Queryable(self.it, query_funcs, **kwargs)

    def where(self, query_func=None, **kwargs):
        if not query_func and not kwargs:
            return self
        new_funcs = self.query_funcs + ((query_func,) if query_func else ())
        new_kwargs = copy(self.kwargs)
        new_kwargs.update(kwargs)
        return self._new(new_funcs, new_kwargs)

    @staticmethod
    def chain(*its):
        return ChainedQueryable(*its)


class ChainedQueryable(Queryable):

    def __init__(self, *its):
        self.its = its
        self.it = itertools.chain(*its)

    def __next__(self):
        return next(self.it)

    def where(self, query_func=None, **kwargs):
        its = []
        for it in self.its:
            its.append(it.where(query_func, **kwargs))
        return ChainedQueryable(*its)



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

import collections
import errno
import os
import sys
import zipfile

SETUP_DIR = os.path.abspath(os.path.dirname(__file__))
VERSION = '0.1.0'


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ['--pep8']

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        fetch_test_data()

        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


class ExtraRequires(collections.Mapping):
    suffix = '-requirements.txt'

    def __getitem__(self, key):
        return get_requires(key + self.suffix)

    def __iter__(self):
        for filename in os.listdir(SETUP_DIR):
            if (os.path.isfile(filename) and filename.endswith(self.suffix)):
                yield filename[:-len(self.suffix)]

    def __len__(self):
        return sum(1 for _ in self)


def get_requires(filename='requirements.txt'):
    with open(os.path.join(SETUP_DIR, filename), 'r') as f:
        for line in f:
            line = line.strip()
            if line.startswith('#') or not line:
                continue
            elif '#' in line:
                yield line.split('#', 1)[0]
            else:
                yield line


def fetch_test_data():
    """
    Fetches test data if it's missing.
    """

    data_dir = os.path.join(SETUP_DIR, 'tests', 'data')

    # if tests/data exists, is empty, and .gitmodules exists, run the
    # appropriate git-submodule(1) command
    if (os.path.isdir(data_dir) and not os.listdir(data_dir) and
            os.path.isfile(os.path.join(SETUP_DIR, '.gitmodules'))):
        import subprocess
        subprocess.call(['git', 'submodule', 'update',
                         '--init', '--recursive'])
    # if tests/data either doesn't exist or is empty, download the
    # busbus-test-data zip from GitHub and unpack it
    else:
        try:
            os.mkdir(data_dir)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

        if not os.listdir(data_dir):
            # at this point requests and six are already installed
            import requests
            import six

            resp = requests.get('https://github.com/spaceboats/busbus-test-'
                                'data/archive/v{0}.zip'.format(VERSION))
            with zipfile.ZipFile(six.BytesIO(resp.content)) as z:
                prefix = 'busbus-test-data-{0}/'.format(VERSION)
                for filename in z.namelist():
                    if filename == prefix or not filename.startswith(prefix):
                        continue
                    filename = filename[len(prefix):]
                    if filename.endswith('/'):
                        os.mkdir(os.path.join(data_dir, filename))
                    else:
                        with open(os.path.join(data_dir, filename), 'wb') as f:
                            f.write(z.read(prefix + filename))


with open('README.rst') as f:
    long_description = f.read().replace(' :bus:', '')

setup(
    name='busbus',
    version=VERSION,
    author='Ian Weller',
    author_email='ianweller@buttslol.net',
    description='A platform for working with public transit data',
    long_description=long_description,
    license='MIT',
    url='https://github.com/spaceboats/busbus',
    packages=find_packages(exclude=['tests*']),
    package_data={
        'busbus.provider': ['*.sql'],
    },
    install_requires=list(get_requires()),
    tests_require=list(get_requires('dev-requirements.txt')),
    extras_require=ExtraRequires(),
    cmdclass={
        'test': PyTest,
    },
)



import busbus
from busbus.queryable import Queryable
from busbus import util

from abc import ABCMeta, abstractmethod, abstractproperty
import arrow
import collections
import datetime
import heapq
import six


@six.add_metaclass(ABCMeta)
class ArrivalGeneratorBase(util.Iterable):

    def __init__(self, provider, stops, routes, start, end):
        self.provider = provider
        self.stops = stops
        self.routes = routes
        self.start = (arrow.now() if start is None
                      else arrow.get(start)).to(provider._timezone)
        self.end = (self.start.replace(hours=3) if end is None
                    else arrow.get(end)).to(provider._timezone)
        self.it = None

    @abstractproperty
    def realtime(self):
        """
        True if this generator's Arrival objects can represent realtime data,
        otherwise False.

        Setting this to True doesn't necessarily mean that all the generated
        arrivals will be realtime.

        You don't have to define a property function -- just set the class
        property during the definition, like this:

        class MyArrivalGenerator(ArrivalGeneratorBase):
            realtime = False
        """

    @abstractmethod
    def _build_iterable(self):
        """
        Build an iterator that provides arrivals.

        This is done in a separate function rather than in __init__ so that
        this generator can be lazily evaluated.
        """

    def __next__(self):
        if self.it is None:
            self.it = self._build_iterable()
        return next(self.it)


class ArrivalQueryable(Queryable):

    def __init__(self, provider, arrival_gens, query_funcs=None, **kwargs):
        self.provider = provider

        if isinstance(arrival_gens, collections.Iterable):
            self.arrival_gens = tuple(arrival_gens)
        else:
            self.arrival_gens = (arrival_gens,)

        if 'realtime' in kwargs:
            realtime = bool(kwargs['realtime'])
        else:
            realtime = any(gen.realtime for gen in self.arrival_gens)

        if 'stop' in kwargs:
            stops = [kwargs.pop('stop')]
        elif 'stop.id' in kwargs:
            stop = provider.get(busbus.Stop, kwargs.pop('stop.id'), None)
            stops = [] if stop is None else [stop]
        else:
            stops = None

        if 'route' in kwargs:
            routes = [kwargs.pop('route')]
        elif 'route.id' in kwargs:
            route = provider.get(busbus.Route, kwargs.pop('route.id'), None)
            routes = [] if route is None else [route]
        else:
            routes = None

        for attr in ('start_time', 'end_time'):
            if attr in kwargs:
                if isinstance(kwargs[attr], datetime.datetime):
                    kwargs[attr] = arrow.Arrow.fromdatetime(kwargs[attr])
                elif isinstance(kwargs[attr], datetime.date):
                    kwargs[attr] = arrow.Arrow.fromdate(kwargs[attr])
        start = kwargs.pop('start_time', None)
        end = kwargs.pop('end_time', None)

        it = heapq.merge(*[gen(provider, stops, routes, start, end)
                           for gen in self.arrival_gens
                           if gen.realtime == realtime])
        super(ArrivalQueryable, self).__init__(it, query_funcs, **kwargs)

    def _new(self, query_funcs, kwargs):
        return ArrivalQueryable(self.provider, self.arrival_gens,
                                query_funcs, **kwargs)
